/* vi:set shiftwidth=2 cindent tabstop=2: */
/*
 * vxtools - Tools for accessing Veritas Journaled FileSystem (VxFS)
 * Copyright (c) 1999 Martin Hinner <mhi@penguin.cz>
 *
 * vxls.c: List a directory contents
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * $Id: vxls.c,v 0.3 1999/12/14 19:04:35 mhi Exp root $
 */
#include <stdio.h>
#include <stdlib.h>
#include <getopt.h>
#include <string.h>
#include <sys/time.h>

#include <libvxfs.h>

char *usage =
"Usage: %s [-1FRadhil] [name...]\n"
"\n"
"Report bugs to <mhi@penguin.cz>\n";

int printinode;
int classify;
int recursive;
int allfiles;
int longlist;

void
listdir (struct vxfs *fs, char *dirname, char *name)
{
	int ino;
	struct vxfs_inode *inode, *dino;
	struct vxdirent **files, *dirent;
	char *rpath1, *rpath2;				/* Paths for recursive ls */
	int numdir;
	int i;
	char *symlink;

	if (name)
		printf ("%s:\n", name);

	ino = vxlookup (fs, dirname);
	if (ino < 0)
		{
			printf ("%s: No such file or directory\n", dirname);
			return;
		}
	inode = vxiget (fs, ino);

	if (inode == NULL)
		{
			printf ("%s: Can't get inode %u\n", dirname, ino);
			return;
		}

	numdir = vxscandir (fs, inode, &files, NULL, vxalphasort);
	if (numdir == -1)
		{
			printf ("%s: vxscandir failed\n", dirname);
			return;
		}

	for (i = 0; i < numdir; i++)
		{
			dirent = files[i];
			/* Skip dot files if not '-a' */
			if (dirent->name[0] == '.' && !allfiles)
				continue;

			dino = vxiget (fs, dirent->inode);
			if (dino == NULL)
				{
					fprintf (stderr, "Can't get inode %u\n", dirent->inode);
					return;
				}

			if (printinode)
				printf ("%8u ", dirent->inode);

			if (longlist)
				{
					if (VXFS_ISDIR (dino))
						printf ("d");
					else if (VXFS_ISBLK (dino))
						printf ("b");
					else if (VXFS_ISCHR (dino))
						printf ("c");
					else if (VXFS_ISLNK (dino))
						printf ("l");
					else if (VXFS_ISSOC (dino))
						printf ("s");
					else if (VXFS_ISFIFO (dino))
						printf ("f");
					else
						printf ("-");

					printf ("%c%c%c%c%c%c%c%c%c",
									dino->mode & 0400 ? 'r' : '-',
									dino->mode & 0200 ? 'w' : '-',
									dino->mode & 0100 ? 'x' : '-',
									dino->mode & 0040 ? 'r' : '-',
									dino->mode & 0020 ? 'w' : '-',
									dino->mode & 0010 ? 'x' : '-',
									dino->mode & 0004 ? 'r' : '-',
									dino->mode & 0002 ? 'w' : '-',
									dino->mode & 0001 ? 'x' : '-');

					printf (" %4u %-4u %-4u ",
									dino->nlink, dino->uid, dino->gid);

					if (VXFS_ISCHR (dino) || VXFS_ISBLK (dino))
						{
							printf ("%3u, %3u ", dino->ftarea.rdev >> 16,
											dino->ftarea.rdev & 0xFFFF);
						}
					else
						printf ("%8lu ", (unsigned long long) dino->size);

					printf ("%-24.24s ", ctime (&dino->mtime));
				}

			printf ("%s", dirent->name);
			if (classify)
				{
					if (VXFS_ISDIR (dino))
						printf ("/");
					if (VXFS_ISLNK (dino) && !longlist)
						printf ("@");
					if (VXFS_ISSOC (dino))
						printf ("=");
					if (VXFS_ISFIFO (dino))
						printf ("|");
					if (VXFS_ISREG (dino) && ((dino->mode & 0500 == 0500) ||
											(dino->mode & 050 == 050) || (dino->mode & 05 == 05)))
						printf ("*");
				}

			if (VXFS_ISLNK (dino) && longlist)
				{
					symlink = vxiread (fs, dino, 0, 1);
					if (symlink)
						{
							printf (" -> %*.*s", (int) (dino->size > fs->bsize ? fs->bsize : dino->size),
											(int) (dino->size > fs->bsize ? fs->bsize : dino->size), symlink);
							free (symlink);
						}
				}
			printf ("\n");

			vxiput (dino);
		}

	if (recursive)
		{
			printf ("\n");

			for (i = 0; i < numdir; i++)
				{
					dirent = files[i];

					if (strcmp (dirent->name, ".."))
						{
							dino = vxiget (fs, dirent->inode);
							if (dino == NULL)
								{
									fprintf (stderr, "Can't get inode %u\n", dirent->inode);
									return;
								}

							if (VXFS_ISDIR (dino))
								{
									rpath1 = (char *) malloc (strlen (dirname) + strlen (dirent->name) + 2);
									strcpy (rpath1, dirname);
									strcat (rpath1, "/");
									strcat (rpath1, dirent->name);
									if (name)
										{
											rpath2 = (char *) malloc (strlen (name) + strlen (dirent->name) + 2);
											strcpy (rpath2, name);
											if (rpath2[strlen (rpath2) - 1] != '/')
												strcat (rpath2, "/");
											strcat (rpath2, dirent->name);
										}
									else
										rpath2 = strdup (dirent->name);

									listdir (fs, rpath1, rpath2);		/* Recursive ls */

									free (rpath1);
									free (rpath2);
								}
							vxiput (dino);
						}
				}												/* for () */
		}														/* if (recursive) */

	for (i = 0; i < numdir; i++)
		free (files[i]);
	free (files);

	vxiput (inode);
}


int
main (int argc, char **argv)
{
	int ch;
	char *argv0 = argv[0];
	char device[0x100], cwd[0x100];
	int subpart;
	char path[0x100];
	int printname;
	struct vxfs *fs;

	vxverboseon ();

	printinode = 0;
	classify = 0;
	recursive = 0;
	allfiles = 0;
	longlist = 0;

	while ((ch = getopt (argc, argv, "adhilFR1")) != -1)
		{
			switch (ch)
				{
				case '1':
					/* Ignore for compatibility */
					break;
				case 'a':
					allfiles = 1;
					break;
				case 'd':
					vxdebugon ();
					break;
				case 'h':
					fprintf (stderr, usage, argv0);
					exit (0);
				case 'i':
					printinode = 1;
					break;
				case 'l':
					longlist = 1;
					break;
				case 'F':
					classify = 1;
					break;
				case 'R':
					recursive = 1;
					break;
				default:
					fprintf (stderr, usage, argv0);
					exit (1);
				}
		}

	argc -= optind;
	argv += optind;

	if (vxreadcwd (device, &subpart, cwd) < 0)
		{
			fprintf (stderr, "%s: No VxFS filesystem mounted, use vxmount(8)\n", argv0);
			exit (1);
		}
	if ((fs = vx_open (device, subpart)) == NULL)
		{
			fprintf (stderr, "%s: Can't mount VxFS %s\n", argv0, device);
			exit (1);
		}
	printname = argc > 1;

	if (argc)
		{
			while (argc--)
				{
					if (argv[0][0] == '/')
						strcpy (path, argv[0]);
					else
						{
							strcpy (path, cwd);
							strcat (path, "/");
							strcat (path, argv[0]);
						}
					listdir (fs, path, printname ? argv[0] : 0);
					argv++;
				}
		}
	else
		{
			listdir (fs, cwd, 0);
		}

	return 0;
}
